
/* Chrysalide - Outil d'analyse de fichiers binaires
 * call.c - encadrement des appels de routine
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "call.h"


#include <malloc.h>
#include <string.h>


#include "../expression-int.h"



/* Définition d'un appel à une routine quelconque (instance) */
struct _GRoutineCall
{
    GDecExpression parent;                  /* A laisser en premier        */

    GBinRoutine *routine;                   /* Routine sollicitée          */

    GDecInstruction **args;                 /* Arguments à associer        */
    size_t count;                           /* Nombre d'arguments présents */

};


/* Définition d'un appel à une routine quelconque (classe) */
struct _GRoutineCallClass
{
    GDecExpressionClass parent;             /* A laisser en premier        */

};



/* Initialise la classe des appels à une routine quelconque. */
static void g_routine_call_class_init(GRoutineCallClass *);

/* Initialise une instance d'appel à une routine quelconque. */
static void g_routine_call_init(GRoutineCall *);

/* Visite un ensemble hiérarchique d'instructions décompilées. */
static bool g_routine_call_visit(GRoutineCall *, dec_instr_visitor_cb, DecInstrVisitFlags, void *);

/* Remplace une instruction décompilée par une autre. */
static bool g_routine_call_replace(GRoutineCall *, GDecInstruction *, GDecInstruction *);

/* Imprime pour l'écran un version humaine d'une expression. */
static GBufferLine *g_routine_call_print(const GRoutineCall *, GCodeBuffer *, GBufferLine *, GLangOutput *);



/* Indique le type défini pour un appel à une routine quelconque. */
G_DEFINE_TYPE(GRoutineCall, g_routine_call, G_TYPE_DEC_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des appels à une routine quelconque.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_routine_call_class_init(GRoutineCallClass *klass)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'appel à une routine quelconque.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_routine_call_init(GRoutineCall *expr)
{
    GDecInstruction *instr;                 /* Autre version de l'objet    */

    instr = G_DEC_INSTRUCTION(expr);

    instr->visit = (dec_instr_visit_fc)g_routine_call_visit;
    instr->replace = (dec_instr_replace_fc)g_routine_call_replace;
    instr->print = (dec_instr_print_fc)g_routine_call_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : routine = routine dont il est fait appel.                    *
*                                                                             *
*  Description : Exprime un appel à une routine quelconque.                   *
*                                                                             *
*  Retour      : Expression mise en place.                                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *g_routine_call_new(GBinRoutine *routine)
{
    GRoutineCall *result;              /* Expression à retourner      */

    result = g_object_new(G_TYPE_ROUTINE_CALL, NULL);

    result->routine = routine;

    return G_DEC_INSTRUCTION(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call     = première instruction à venir visiter.             *
*                callback = procédure à appeler à chaque instruction visitée. *
*                flags    = moments des appels à réaliser en retour.          *
*                data     = données quelconques associées au visiteur.        *
*                                                                             *
*  Description : Visite un ensemble hiérarchique d'instructions décompilées.  *
*                                                                             *
*  Retour      : true si le parcours a été jusqu'à son terme, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_routine_call_visit(GRoutineCall *call, dec_instr_visitor_cb callback, DecInstrVisitFlags flags, void *data)
{
    bool result;                            /* Bilan à retourner           */
    size_t i;                               /* Boucle de parcours          */

    result = true;

    for (i = 0; i < call->count && result; i++)
        result = _g_dec_instruction_visit(call->args[i], G_DEC_INSTRUCTION(call),
                                          callback, flags, data);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = première instruction à venir ausculter.               *
*                old  = instruction décompilée à venir remplacer.             *
*                new  = instruction décompilée à utiliser dorénavant.         *
*                                                                             *
*  Description : Remplace une instruction décompilée par une autre.           *
*                                                                             *
*  Retour      : true si un remplacement a été effectué, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_routine_call_replace(GRoutineCall *call, GDecInstruction *old, GDecInstruction *new)
{
    bool result;                            /* Bilan à retourner           */
    size_t i;                               /* Boucle de parcours          */

    result = false;

    for (i = 0; i < call->count; i++)
    {
        if (call->args[i] == old)
        {
            g_object_unref(G_OBJECT(call->args[i]));
            g_object_ref(G_OBJECT(new));
            call->args[i] = new;

            result = true;

        }
        else
            result |= g_dec_instruction_replace(call->args[i], old, new);

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call   = expression à transcrire en version humaine.         *
*                buffer = tampon où doit se réaliser l'insertion.             *
*                line   = ligne d'impression prête à emploi ou NULL.          *
*                output = langage de programmation de sortie.                 *
*                                                                             *
*  Description : Imprime pour l'écran un version humaine d'une expression.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GBufferLine *g_routine_call_print(const GRoutineCall *call, GCodeBuffer *buffer, GBufferLine *line, GLangOutput *output)
{
    GBufferLine *result;                    /* Ligne à retourner           */
    const char *name;                       /* Désignation de la routine   */
    size_t i;                               /* Boucle de parcours          */

    name = g_binary_routine_get_name(call->routine);
    g_buffer_line_append_text(line, BLC_ASSEMBLY_HEAD, name, strlen(name), RTT_RAW, NULL);

    g_buffer_line_append_text(line, BLC_ASSEMBLY_HEAD, "(", 1, RTT_PUNCT, NULL);

    if (call->count > 0)
    {
        result = g_dec_instruction_print(call->args[0], buffer, line, output);

        for (i = 1; i < call->count; i++)
        {
            g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, ",", 1, RTT_PUNCT, NULL);
            g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, " ", 1, RTT_RAW, NULL);

            g_dec_instruction_print(call->args[i], buffer, result, output);

        }

    }
    else result = line;

    g_buffer_line_append_text(result, BLC_ASSEMBLY_HEAD, ")", 1, RTT_PUNCT, NULL);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = expression d'appel à mettre à jour.                   *
*                arg  = nouvel argument à associer à l'appel.                 *
*                                                                             *
*  Description : Enregistre un nouvel argument pour l'appel de routine.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_routine_call_add_arg(GRoutineCall *call, GDecInstruction *arg)
{
    call->args = (GDecInstruction **)realloc(call->args,
                                             ++call->count * sizeof(GDecInstruction *));

    call->args[call->count - 1] = arg;

    /* TODO : synchroniser avec la routine (cf. constructeurs construits à la volée) */

}
