
/* Chrysalide - Outil d'analyse de fichiers binaires
 * return.c - représentation des ordres de retour
 *
 * Copyright (C) 2010-2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "return.h"


#include "../expression-int.h"



/* Définition d'un ordre de retour (instance) */
struct _GReturnExpression
{
    GDecExpression parent;                  /* A laisser en premier        */

    GDecExpression *payload;                /* Eventuel paquet associé     */

};


/* Définition d'un ordre de retour (classe) */
struct _GReturnExpressionClass
{
    GDecExpressionClass parent;             /* A laisser en premier        */

};



/* Initialise la classe des ordres de retour. */
static void g_return_expression_class_init(GReturnExpressionClass *);

/* Initialise une instance d'ordre de retour. */
static void g_return_expression_init(GReturnExpression *);

/* Visite un ensemble hiérarchique d'instructions décompilées. */
static bool g_return_expression_visit(GReturnExpression *, dec_instr_visitor_cb, DecInstrVisitFlags, void *);

/* Remplace une instruction décompilée par une autre. */
static bool g_return_expression_replace(GReturnExpression *, GDecInstruction *, GDecInstruction *);

/* Imprime pour l'écran un version humaine d'une expression. */
static GBufferLine *g_return_expression_print(const GReturnExpression *, GCodeBuffer *, GBufferLine *, GLangOutput *);



/* Indique le type défini pour un ordre de retour. */
G_DEFINE_TYPE(GReturnExpression, g_return_expression, G_TYPE_DEC_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des ordres de retour.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_return_expression_class_init(GReturnExpressionClass *klass)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'ordre de retour.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_return_expression_init(GReturnExpression *expr)
{
    GDecInstruction *instr;                 /* Autre version de l'objet    */

    instr = G_DEC_INSTRUCTION(expr);

    instr->visit = (dec_instr_visit_fc)g_return_expression_visit;
    instr->replace = (dec_instr_replace_fc)g_return_expression_replace;
    instr->print = (dec_instr_print_fc)g_return_expression_print;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : payload = éventuelle expression à associer au retour.        *
*                                                                             *
*  Description : Constitue un ordre de retour vers l'appelant.                *
*                                                                             *
*  Retour      : Expression mise en place.                                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GDecInstruction *g_return_expression_new(GDecExpression *payload)
{
    GReturnExpression *result;              /* Expression à retourner      */

    result = g_object_new(G_TYPE_RETURN_EXPRESSION, NULL);

    result->payload = payload;

    return G_DEC_INSTRUCTION(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr     = première instruction à venir visiter.             *
*                callback = procédure à appeler à chaque instruction visitée. *
*                flags    = moments des appels à réaliser en retour.          *
*                data     = données quelconques associées au visiteur.        *
*                                                                             *
*  Description : Visite un ensemble hiérarchique d'instructions décompilées.  *
*                                                                             *
*  Retour      : true si le parcours a été jusqu'à son terme, false sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_return_expression_visit(GReturnExpression *expr, dec_instr_visitor_cb callback, DecInstrVisitFlags flags, void *data)
{
    bool result;                            /* Bilan à retourner           */

    if (expr->payload)
        result = _g_dec_instruction_visit(G_DEC_INSTRUCTION(expr->payload), G_DEC_INSTRUCTION(expr),
                                          callback, flags, data);
    else
        result = false;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = première instruction à venir ausculter.               *
*                old  = instruction décompilée à venir remplacer.             *
*                new  = instruction décompilée à utiliser dorénavant.         *
*                                                                             *
*  Description : Remplace une instruction décompilée par une autre.           *
*                                                                             *
*  Retour      : true si un remplacement a été effectué, false sinon.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_return_expression_replace(GReturnExpression *expr, GDecInstruction *old, GDecInstruction *new)
{
    bool result;                            /* Bilan à retourner           */

    if (expr->payload)
    {
        if (expr->payload == G_DEC_EXPRESSION(old))
        {
            g_object_unref(G_OBJECT(expr->payload));
            g_object_ref(G_OBJECT(new));
            expr->payload = G_DEC_EXPRESSION(new);

            result = true;

        }
        else
            result = g_dec_instruction_replace(G_DEC_INSTRUCTION(expr->payload), old, new);

    }
    else
        result = false;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr   = expression à transcrire en version humaine.         *
*                buffer = tampon où doit se réaliser l'insertion.             *
*                line   = ligne d'impression prête à emploi ou NULL.          *
*                output = langage de programmation de sortie.                 *
*                                                                             *
*  Description : Imprime pour l'écran un version humaine d'une expression.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GBufferLine *g_return_expression_print(const GReturnExpression *expr, GCodeBuffer *buffer, GBufferLine *line, GLangOutput *output)
{
    GBufferLine *result;                    /* Ligne à retourner           */

    g_buffer_line_append_text(line, BLC_ASSEMBLY_HEAD, "return", 6, RTT_KEY_WORD, NULL);

    if (expr->payload != NULL)
    {
        g_buffer_line_append_text(line, BLC_ASSEMBLY_HEAD, " ", 1, RTT_RAW, NULL);

        result = g_dec_instruction_print(G_DEC_INSTRUCTION(expr->payload),
                                         buffer, line, output);

    }
    else result = line;

    return result;

}
